/*
 * xds510 simulator for Wine
 * Copyright (c) 2001 Blaise Gassend (blaise.gassend@m4x.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. This program is
 * distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. You should have received a
 * copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 * */

// This code assumes at least 32-bit ints as does wine.

#define PAR_ADDR (0x378)

#include <stdio.h>

#include "xds510.h"
#include "xds510_io.h"
#include "debugtools.h"

DECLARE_DEBUG_CHANNEL(xds510);
DEFAULT_DEBUG_CHANNEL(xds510);

#define BTAB "\t\t\t\t"
 
#define INTERFACE_VERSION (6)
static int interface_interrupt;
//static int read800;
static int interface_reset;
static int counter1;
static int counter1_update;
static int counter1_capture;
static int active_command;

#define NUM_LOW_REGISTERS (0xC)
#define NUM_STATUS_REGISTERS (4)
static int low_register_val[NUM_LOW_REGISTERS];
static int status_register[NUM_STATUS_REGISTERS];

static int fifo_buffer;
static unsigned int read_buffer;
static int read_buffer_pos;
static unsigned int write_buffer;
static int write_buffer_pos;

#define TAP_TEST_LOGIC_RESET (0)
#define TAP_RUN_TEST_IDLE (2)
#define TAP_SHIFT_DR (4)
#define TAP_SHIFT_IR (5)
#define TAP_PAUSE_DR (6)
#define TAP_PAUSE_IR (7)

static int tap_state;

//static FILE *debugf; // ####
//static int outdata, indata, indatap, outdatap, databitsp;

int get_outbit()
{
  int outbit;
  
  switch (low_register_val[4] & 0x0300)
  {
    default:
    case 0x0000:
      outbit = 0;
      break;
      
    case 0x0100:
      outbit = 1;
      break;
      
    case 0x0200:
      outbit = fifo_buffer & 1;
      break;

    case 0x0300:
      outbit = write_buffer & 1;
      break;
  }
  
  return outbit;
}

/*static void tap_log(int tms, int tdo, int tdi)
{
  void printbuff(unsigned char *buff, int len)
  {
    int i;
    int bytes = len / 8;
    int bits = len % 8;
    int mask = 1 << bits;
    
    for (i = 0; i < bytes; i++)
    {
      fprintf(debugf, "%02x", (int) buff[i]);
    }

    if (bits > 0)
    {
      buff[bytes] &= mask - 1;
      fprintf(debugf, "%0*x", (bits + 3) / 4, buff[bytes]);
    }
  }
  
  static unsigned char inbuff[256];
  static unsigned char outbuff[256];
  static int buffp = 0;
  static int state = 0; // 0 reset, 1 run, then DR top to bot, ir top to bot.
  static int trans[2][16] = 
  { { 1, 1, 3, 4, 4, 6, 6, 4, 1, 10, 11, 11, 13, 13, 11, 1 },
    { 0, 2, 9, 5, 5, 8, 7, 8, 2, 0, 12, 12, 15, 14, 15, 2 } };

  if (state == 4 || state == 11) // Shifting
  {
    int byte = buffp / 8;
    int bit = buffp % 8;
    int mask = 1 << bit;
    if (buffp < 2048)
    {
      inbuff[byte] = (inbuff[byte] & (mask - 1)) | (tdi ? mask : 0);
      outbuff[byte] = (outbuff[byte] & (mask - 1)) | (tdo ? mask : 0);
      buffp++;
    }
  }
  
  state = trans[tms][state & 0xF];

  switch (state)
  {
    case 0: // reset
      fputs("RESET ", debugf);
      break;

    case 1: // run
      fputs("RUN ", debugf);
      break;

    case 8: // update dr
      fprintf(debugf, "\n              DRU %i ", buffp);
      printbuff(outbuff, buffp);
      fputs(" -> ", debugf);
      printbuff(inbuff, buffp);
      fputc(' ', debugf);
      fflush(debugf);
      buffp = 0;
      break;
       
    case 15: // update ir
      fprintf(debugf, "\nIRU %i ", buffp - 16);
      printbuff(outbuff + 2, buffp - 16);
      fputs(" -> ", debugf);
      printbuff(inbuff, buffp - 16);
      fputc(' ', debugf);
      fflush(debugf);
      buffp = 0;
      break;
  }
}
*/
static int io_cycle(int tms)
{
  int tdo = get_outbit();
  int tdi = xds510_io_write_cycle(tdo, tms);

//  tap_log(tms, tdo, tdi);
  
  return tdi;
}

static void XDS510_reset()
{
  int i;
  
  TRACE("XDS510_reset\n");

  counter1 = 0;
  counter1_update = 0;
  counter1_capture = 0;
  interface_interrupt = 0;
//  read800 = 0;
  active_command = 0;
  
  fifo_buffer = 0xFFFFFFFF;
  read_buffer = 0;
  read_buffer_pos = 0;
  write_buffer = 0;
  write_buffer_pos = 0;

/*  for (i = 0; i < 6; i++)
  {
    io_cycle(1); 
  }
  io_cycle(0); */
  tap_state = TAP_TEST_LOGIC_RESET;
  
  for (i = 0; i < NUM_LOW_REGISTERS; i++)
    low_register_val[i] = 0;    
  
  for (i = 0; i < NUM_STATUS_REGISTERS; i++)
    status_register[i] = 0;    

  TRACE("returning\n\n");
}

void XDS510_init()
{
  TRACE("XDS510_init\n");
  
  if (xds510_io_init(PAR_ADDR))
  {
    MESSAGE("\n\n***** Root privileges necessary to access XDS510.\n\n");
  }
  else
  {
    MESSAGE("\n\n***** XDS510 on port 0x%03x.\n\n", PAR_ADDR);
  }
  
//  if (debugf == NULL)
//    debugf = /*stderr; /*/fopen("debug.txt", "w");
  
  xds510_io_write_trst(0);
  io_cycle(1);
  xds510_io_write_trst(1);
  io_cycle(1);
  
  XDS510_reset();
}

static void go_tap_state(int state)
{
  // The order is RESET, RUN, S-DR, S-IR, P-DR, P-IR
  
  static const char *transitions[6][6] = { // [from][to]
    { "11111", "0",   "0100",  "01100",  "01010",  "011010"  },
    { "11111", "0",   "100",   "1100",   "1010",   "11010"   },
    { "11111", "110", "0",     "111100", "10",     "1111010" },
    { "11111", "110", "11100", "0",      "111010", "10"      },
    { "11111", "110", "10",    "111100", "0",      "1111010" },
    { "11111", "110", "11100", "10",     "111010", "0"    }  };
  int i;

  int state_id(int s)
  {
    s &= 0x7;
    if (s < 4) 
      return s / 2;
    else 
      return s - 2;
  }
  
  const char *trans = transitions[state_id(tap_state)][state_id(state)];
  
//  printf("Goto Tap %i From %i\n", state, tap_state);
  
  while (*trans)
  {
    io_cycle(*trans - '0');
    trans++;
  }

  if ((state & 0x6) == 2)
    for (i = 0; i < 10; i++)
      io_cycle(0);  

  tap_state = state & 0x7;
}

static void go_tap_state2(int state)
{
  // The order is RESET, RUN, S-DR, S-IR, P-DR, P-IR
  
  static const char *transitions[6][6] = { // [from][to]
    { "11111", "0",   "0100",  "01100",  "01010",  "011010"  },
    { "11111", "0",   "100",   "1100",   "1010",   "11010"   },
    { "11111", "110", "0",     "111100", "10",     "1111010" },
    { "11111", "110", "11100", "0",      "111010", "10"      },
    { "11111", "110", "10",    "111100", "111010", "1111010" },
    { "11111", "110", "11100", "10",     "111010", "1111010" }  };
  int i;
  
  int state_id(int s)
  {
    s &= 0x7;
    if (s < 4) 
      return s / 2;
    else 
      return s - 2;
  }
  
  const char *trans = transitions[state_id(tap_state)][state_id(state)];
  
//  printf("Goto Tap 2 %i From %i\n", state, tap_state);
  
  while (*trans)
  {
    io_cycle(*trans - '0');
    trans++;
  }

  if ((state & 0x6) == 2)
    for (i = 0; i < 10; i++)
      io_cycle(0);  

  tap_state = state & 0x7;
}

static void minor_command(int value)
{
  int cmd = (value >> 12) & 0xF;

  switch (cmd)
  {
    case 4:
    case 5:
      if (value & 0x0080)
      {
/*	if (low_register_val[0xB] >> 12 == 3)
	{
	  fprintf(debugf, "i %0*x o %0*x\nExecute %04X, count = %i ", 
	    (databitsp + 3) / 4, indata & ~(-1 << (indatap - 16)), 
	    (databitsp + 3) / 4, outdata >> 16, 
	    low_register_val[0xB], counter1 - 15);
	  fflush(debugf);
	  outdata = 0;
	  indata = 0;
	  indatap = 0;
	  outdatap = 0;
	  databitsp = counter1 - 15;
	}*/
	active_command = 1;
	TRACE("Just activated a command.\n");
      }
      if (value & 0xF7F)
      {
	FIXME("Unsupported bits in minor command 4\n");
      }
      break;
    
    case 6:
      if (value & 0x001)
      {
        counter1 = counter1_update;
	TRACE("counter1 loaded with : %08x\n", counter1);
      }
      
      if (value & 0x004) 
      {
	counter1_capture = counter1;
	TRACE("counter1 captured with : %08x\n", counter1);
      }
	
      if (value & 0x040) 
      {
	status_register[0] = 0xF; // ####
	status_register[1] = (active_command ? 0x8000 : 0);
	status_register[2] = tap_state & 0x7;
	if (read_buffer_pos >= 16) status_register[2] |= 0x1000;
	if (read_buffer_pos >= 32) status_register[2] |= 0x2000;
	if (write_buffer_pos > 0) status_register[2] |= 0x0100;
	if (write_buffer_pos > 16) status_register[2] |= 0x0200;
	status_register[3] = 0;
      }
      
      if (value & (0x002 | 0x030 | 0xF88))
      {
	FIXME("Unsupported bits in minor command 6\n");
      }
      break;

    case 7:
      if (value & 0x0001)
      {
	XDS510_reset();
      }

      if (value & 0x0FFE)
      {
	FIXME("Unsupported bits in minor command 7\n");
      }
      
      break;
    
    default:
      FIXME(BTAB"Unknown minor command %04x\n", value);
  }

  TRACE(BTAB"minor_command(%04x)\n", value);
}

static int scannable_bits()
{
  int rslt = counter1 + 1;
  if (rslt > 32 - read_buffer_pos && (low_register_val[4] & 0x0400)) 
  {
    rslt = 32 - read_buffer_pos;
  }
  if (rslt > write_buffer_pos && 
      (low_register_val[4] & 0x0300) == 0x0300) 
  {
    rslt = write_buffer_pos;
  }
  return rslt;
}

static void shift_bit(int tdi0)
{
  if (!(low_register_val[4] & 0x0080))
  {
    tdi0 = get_outbit();
  }

  TRACE("s4 : %04x tdi : %i w : %08x f : %04x r : %08x\n", 
      low_register_val[4], tdi0, write_buffer, fifo_buffer, read_buffer);
  
  if (low_register_val[4] & 0x0400) 
  {
    read_buffer = (read_buffer & ~(-1 << read_buffer_pos)) | 
      ((fifo_buffer & 1) <<  read_buffer_pos);
    read_buffer_pos++;
  }

  fifo_buffer >>= 1;
  {
    int mask = (low_register_val[4] & 0x0020) ? 0x80000000 : 0x8000;
    if (tdi0)
      fifo_buffer |= mask;
    else
      fifo_buffer &= ~mask;
  }
  
  if ((low_register_val[4] & 0x0300) == 0x0300)
  {
    write_buffer >>= 1;
    write_buffer_pos--;
  }
}

static void scan_bits(int pause_state)
{
  int bits = scannable_bits();
  
  if (bits > 0)
  {
    go_tap_state(pause_state);
	
    io_cycle(1); // Switch to scan state
    io_cycle(0);
    
    while (bits > 1)
    {
      counter1--;
      bits--;
      shift_bit(io_cycle(0));
    }

    counter1--;
    shift_bit(io_cycle(1)); // Return to pause state
    io_cycle(0);  
  }
}

static void execute_command()
{
  if (active_command)
  {
    int fcmd = low_register_val[0xB];
    int cmd = (fcmd >> 12) & 0xF;
    
    switch (cmd)
    {
      case 1:
	TRACE("STATE command to TAP state %i.\n", fcmd & 0x7);
	active_command = 0;
	go_tap_state2(fcmd & 0x7);
	if (fcmd & 0x0FF8)
	{
	  FIXME("Unsupported bits in major command 1.\n");
	}
	break;
	
      case 3:
	scan_bits((fcmd & 0x0008) ? TAP_PAUSE_IR : TAP_PAUSE_DR);

	if (fcmd & 0xF80)
	{
	  FIXME("Unsupported bits in major command 3.\n");
	}
	
	if (counter1 < 0)
	{
	  TRACE("SCAN command finished.\n");
	  go_tap_state(fcmd & 0x7);
	  active_command = 0;
	  read_buffer_pos = (read_buffer_pos + 15) & 0xFFF0;
	  write_buffer_pos &= 0xFFF0;
	}
	break;
    
      default:
	active_command = 0;
	FIXME("Tried to execute unsupported major command %04x\n", fcmd);
	break;
    }
  }
}

int XDS510_read(int port)
{
  int retval = 0xFFFF;
  TRACE("XDS510_read(%03x)\n", port);
  
  switch (port)
  {
    case 0x18: // counter1 low
      TRACE("Read counter1 capture low\n");
      retval = counter1_capture & 0xFFFF;
      break;	  
    
    case 0x1A: // counter1 high
      TRACE("Read counter1 capture high\n");
      retval = (counter1_capture >> 16) & 0xFFFF;
      break;	  

    case 0x40C: // Read buffer
      retval = read_buffer & 0xFFFF;  
//      outdata |= retval << outdatap;
//      outdatap += 16;
      read_buffer >>= 16;
      read_buffer_pos -= 16;
      TRACE("Read %04x from read buffer.\n", retval);
      if (read_buffer_pos < 0)
      {
	FIXME("Read buffer underflow.\n");
	read_buffer_pos = 0;
      }
      break;
      
    case 0x800: // The interface word
/*      if (read800)
      {
        fprintf(debugf, "\nread800 %i\n", interface_interrupt);
        interface_interrupt = (interface_interrupt + 1) & 0x1FF;
      }*/
      retval = (((-INTERFACE_VERSION) & 0x7) << 5) | 
	/*(read800 && interface_interrupt == 0) |*/ 0x10 | 
	(!!interface_reset << 8);
/*      if (retval & 0x0001)
        FIXME("interrupt flag %i\n", interface_interrupt);
      read800 = !read800;*/
      break;
      
    default:
      if (port >= 0 && port < 2 * NUM_LOW_REGISTERS)
      {
	retval = low_register_val[port / 2];
      }
      else if (port >= 0x400 && port < 0x400 + 2 * NUM_STATUS_REGISTERS)
      {
	retval = status_register[(port - 0x400) / 2];
      }
      else
      {
        FIXME(BTAB"Unhandled port %03x.\n", port);
      }
      break;
  }
  
  execute_command();
  
/*  if (port != 0x800) 
    read800 = 0;*/
  
  TRACE("returning %04x\n\n", retval);
  return retval;
}

void XDS510_write(int port, int value)
{
  TRACE("XDS510_write(%03x, %04x)\n", port, value);

//  read800 = 0;
  
  if (port >= 0 && port < 2 * NUM_LOW_REGISTERS)
    low_register_val[port / 2] = value & 0xFFFF;
    
  switch (port)
  {
    case 0x0A: //Control 5
      FIXME("Write of %04x to 00A -> emu = %i\n", value, !!(value & 0x0004));
      xds510_io_write_trst(value & 0x0004);
      go_tap_state(tap_state);
      go_tap_state(tap_state);
      break;
    
    case 0x14: // Minor command
      minor_command(value);
      break;

    case 0x8:
    case 0x16: // Major command
//      major_command(value); // #### Useless
      break;

    case 0x18: // Counter1 low
      counter1_update = (counter1_update & 0xFFFF0000) | (value & 0xFFFF);
      TRACE("Write counter1 update low. New value : %08x\n", counter1_update);
      break;

    case 0x1A: // Counter1 high
      counter1_update = (counter1_update & 0xFFFF) | ((value & 0xFFFF) << 16);
      TRACE("Write counter1 update high. New value : %08x\n", counter1_update);	
      break;
    
    case 0x40E: // Write buffer
      write_buffer = (value << write_buffer_pos) | 
	(write_buffer & ~(-1 << write_buffer_pos));
      write_buffer_pos += 16;
      TRACE("Wrote %04x to write buffer.\n", value);
      TRACE("Total bits in buffer : %i\n", write_buffer_pos);
      if (write_buffer_pos > 32)
      {
      	WARN("Write buffer overflow.\n");
	write_buffer_pos = 32;
      }
//      indata |= (value & 0xFFFF) << indatap;
//      indatap += 16;
      break;
	  
    case 0x800: // The interface word
      if (!!(value & 0x100) != interface_reset)
      {
	interface_reset = !!(value & 0x100);
	TRACE("New unknown pin value is %i\n", interface_reset);
//	xds510_io_write_trst(!interface_reset);
      }
      break;
      
    default:
      FIXME(BTAB"Unhandled port %03x. (writing %4x)\n", port, value);
      break;
  }

  execute_command();

  TRACE("returning\n\n");
}
